# BitTorrent Simulator
# written by Fred Wulff (frew@stanford.edu) for CS259
# 3/20/08
# Some of this is pretty hacky, but it proved the feasibility
# of the attack given some conditions, so yay :).

# Assumptions: 
# * Each bucket sent by a good client/seed takes GOOD_ROUND_LENGTH. 
# * Each bucket sent by a bad client takes BAD_ROUND_LENGTH. 
# * All bad clients are able to collaborate, so once a piece is received
#   by one bad client, they effectively all have it.
# * Tit-for-tat is implemented by disconnecting a peer if we try
#   to download from it and fail. In real BT it's implemented via
#   a choking system, the specific's of which depend on the client.
#   The system presented here appears to fall within the spec

import random
import copy

# Number of bystanders and attackers on the network
NUM_GOOD_CLIENTS = 64
NUM_BAD_CLIENTS = 16

# Number of buckets to distribute and number of connections
# /node (note that each node has that many dl connections 
# and that many ul connections)
HASH_TABLE_SIZE = 1024
NUM_CONNECTIONS = 20

# Can the bad clients send fake bucket that doesn't actually
# give the recipient any information
BAD_CLIENT_CAN_POISON = True

# Will the recipient detect it? Currently, if this
# is enabled, the recipient will drop the connection to
# the sender, but won't ban it. This, unfortunately,
# appears to be the state of some mainstream BT
# clients.
BAD_CLIENT_POISON_DETECTABLE = False

GOOD_ROUND_LENGTH = 1
BAD_ROUND_LENGTH = 1

NUM_ROUNDS = 1000

peers = []

def random_elem(rset):
    r = random.randint(0, len(rset) - 1)
    i = 0
    for item in rset:
        if i == r:
            return item
        i+=1
    return None
    
class Client(object):
    def __init__(self, id):
        super(Client, self).__init__()
        self.pieces = set()
        self.id = id
        self.max_connections = NUM_CONNECTIONS
        self.connections = []
        self.num_upload_conn = 0
        self.pieces_to_get = {}
        self.piece_portions = {}
        self.seed = False
        self.remove_bad_peers = True # We don't want to remove leechers
                                     # If we're feeding them garbage
                                     # and we're malicious
        
    # Makes this client a seed at the start
    def set_seed(self):
        self.seed = True
        for i in xrange(HASH_TABLE_SIZE):
            self.pieces.add(i)
    
    # Run to make sure that we have the appropriate number
    # of connections
    def setup_connections(self):
        self.peer_permutations = copy.copy(peers)
        random.shuffle(self.peer_permutations)
        n = min(self.max_connections, len(self.peer_permutations))
        i = len(self.connections)
        while i < n:
            if self.peer_permutations[i] == self or self.peer_permutations[i] in self.connections or self.peer_permutations[i].num_upload_conn > NUM_CONNECTIONS:
                n = min(len(self.peer_permutations), n + 1)
            else:
                self.connections.append(self.peer_permutations[i])
                self.peer_permutations[i].num_upload_conn += 1
            i += 1

    # Do one round of the simulation
    def do_round(self, round_num):
        for peer in peers:
            if not peer.id in self.piece_portions:
                self.piece_portions[peer.id] = 0
        to_remove = []
        for peer in self.connections:
            diff = peer.pieces.difference(self.pieces)
            for key, piece in peer.pieces_to_get.items():
                if piece in diff:
                    diff.remove(piece)
            # print diff
            # We're newly done with a download
            if peer.provides_real_piece() and\
               (self.piece_portions[peer.id] > BAD_ROUND_LENGTH if peer.bad() else self.piece_portions[peer.id] > GOOD_ROUND_LENGTH):
                # print "resetting"
                self.piece_portions[peer.id] = 0
                del self.pieces_to_get[peer.id]
            
            if not self.seed:
                if len(diff) == 0:
                    # print "Removing peer for being boring"
                    # Implement a tit for tat-ish removal algorithm
                    # It appears that BT clients vary on this
                    if self.remove_bad_peers and self in peer.connections:
                        peer.connections.remove(self)
                        self.num_upload_conn += 1
                    to_remove.append(peer)
                else:
                    if not peer.provides_real_piece():
                        while self.piece_portions[peer.id] > BAD_ROUND_LENGTH:
                            if peer.can_detect_fake_piece():
                                to_remove.append(peer)
                            self.piece_portions[peer.id] -= BAD_ROUND_LENGTH
                    else:
                        if self.piece_portions[peer.id] == 0:
                            self.pieces_to_get[peer.id] = random_elem(diff)
                            #if self.id == 25:
                            #    print 'Node %d downloading piece %d from %d to add to (%d)%s' % (self.id, self.pieces_to_get[peer.id], peer.id, len(self.pieces), self.pieces)
                            self.pieces.add(self.pieces_to_get[peer.id])
                    self.piece_portions[peer.id] += 1
                    # print "incrementing"
            if len(self.pieces) == HASH_TABLE_SIZE and not self.seed:
                    print "Node %d became a seed in round %d" % (self.id, round_num)
                    self.seed = True
        for peer in to_remove:
            self.connections.remove(peer)
            peer.num_upload_conn -= 1
    
    def provides_real_piece(self):
        return True
    
    def can_detect_fake_piece(self):
        return False
    
    def bad(self):
        return False
        
# There is only one bad client, which represents
# all of the bad clients (since they're colluding)
class BadClient(Client):
    def __init__(self, id):
        super(BadClient, self).__init__(id)
        self.max_connections = NUM_CONNECTIONS * NUM_BAD_CLIENTS
        self.remove_bad_peers = False
        
    def provides_real_piece(self):
        return not BAD_CLIENT_CAN_POISON
        
    def can_detect_fake_piece(self):
        return BAD_CLIENT_POISON_DETECTABLE
        
    def bad(self):
        return True
        
peers.append(Client(0))
peers[0].set_seed()
for i in xrange(NUM_GOOD_CLIENTS):
    peers.append(Client(i+1))
bad_client = BadClient(NUM_GOOD_CLIENTS+1)
for i in xrange(NUM_BAD_CLIENTS):
    peers.append(bad_client)
    
for i in xrange(NUM_ROUNDS):
    # print "Round %d " % i
    for peer in peers:
        peer.setup_connections()
        peer.do_round(i)
